IMPLEMENTATION MODULE ScrPkT;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    *************************************************************************
    Hardware:   Super Video Graphics Array
    Compiler:   TopSpeed / Stony Brook Modula-2
    -------------------------------------------------------------------------
    A.Iakovlev  Ver  1.0  Derived from generic module RD         October 94
    A.Iakovlev  Ver  1.1  Error correction in StippledLine       February 95
    *************************************************************************
    Hardware dependencies:
    ------ trio-byte/pixel resolution
    ------ sequential pixels positions through all the memory banks
    TopSpeed compiler dependencies:
    ------ predefined type LONGCARD
    ------ single-byte sets
    Stony Brook compiler dependencies:
    ------ predefined type LONGCARD
    ------ single-byte sets
  *)

  IMPORT SrvLow, ScrTyp, Scr;

  TYPE BS = ScrTyp.BS;

  CONST
    BitsPerByte = 8;
    BytesPerTrio = 3;
    BytesPerScrBar = 10000H;

  PROCEDURE Clear(ptr: ScrTyp.BP; count, width: CARDINAL; c: Scr.Pixel);
  (* --- if count = 0, then screen segment is cleared line by line
         to provide correct byte padding in the end of each line *)
    VAR
      line, linesNmb: CARDINAL;
    BEGIN
      IF count = 0 THEN
        linesNmb := VAL(CARDINAL, BytesPerScrBar DIV VAL(LONGCARD, width));
        line := 0;
        LOOP
          SrvLow.TrioFill(ptr.addr, width DIV BytesPerTrio, c);
          IF line = linesNmb - 1 THEN EXIT END;
          INC(ptr.ofs, width);
          INC(line);
        END;
      ELSE
        SrvLow.TrioFill(ptr.addr, count DIV BytesPerTrio, c);
      END;
    END Clear;

  PROCEDURE Pt(ptr: ScrTyp.BP; c: Scr.Pixel; rop: Scr.Function);
    BEGIN
      IF rop = Scr.copy THEN
        SrvLow.TrioPut(ptr.addr, c);
      ELSE
        SrvLow.TrioPutXor(ptr.addr, c);
      END;
    END Pt;

  PROCEDURE PtColor(ptr: ScrTyp.BP): Scr.Pixel;
    BEGIN
      RETURN VAL(Scr.Pixel, SrvLow.TrioGet(ptr.addr));
    END PtColor;

  PROCEDURE SolidLine(ptr: ScrTyp.BP; count: CARDINAL;
            c: Scr.Pixel; rop: Scr.Function);
    BEGIN
      IF rop = Scr.copy THEN
        SrvLow.TrioFill(ptr.addr, count DIV BytesPerTrio, c);
      ELSE
        SrvLow.TrioFillXor(ptr.addr, count DIV BytesPerTrio, c);
      END;
    END SolidLine;

  PROCEDURE StippledLine(ptr: ScrTyp.BP; count: CARDINAL;
            stipplePtr: ScrTyp.BP; stippleBitOfs, stippleWdt: CARDINAL;
            c: Scr.Pixel; rop: Scr.Function);
    VAR
      pixel, bit, ofs0: CARDINAL;
      stippleLarge: BOOLEAN;
      BM: BS;
    BEGIN
      stippleLarge := (stippleWdt > 1);
      ofs0 := stipplePtr.ofs;
      INC(stipplePtr.ofs, stippleBitOfs DIV BitsPerByte);
      BM := stipplePtr.addr^;
      (* -------------------------------- check the stipple *)
      IF (BM # BS{}) OR stippleLarge THEN
        (* ------------- stipple is nonempty or large: draw it *)
        bit := BitsPerByte - 1 - stippleBitOfs MOD BitsPerByte;
        pixel := 0;
        (* -------------------------------------- draw the line *)
        LOOP
          (* ----------------------------- change the pixel color *)
          IF bit IN BM THEN
            IF rop = Scr.copy THEN
              SrvLow.TrioPut(ptr.addr, c);
            ELSE
              SrvLow.TrioPutXor(ptr.addr, c);
            END;
          END;
          (* -------------------------------- check the stop condition *)
          IF pixel = count DIV BytesPerTrio - 1 THEN EXIT END;
          (* ---------- change the stipple bit and (if need) read the stipple *)
          IF bit > 0 THEN
            DEC(bit);
          ELSE
            IF stippleLarge THEN
              IF stipplePtr.ofs - ofs0 < stippleWdt - 1 THEN
                INC(stipplePtr.ofs);
              ELSE
                stipplePtr.ofs := ofs0;
              END;
              BM := stipplePtr.addr^;
            END;
            bit := BitsPerByte - 1;
          END;
          (* ---------- increment the destination offset *)
          INC(ptr.ofs, BytesPerTrio);
          (* ------- increment the pixel counter *)
          INC(pixel);
        END;
      END;
    END StippledLine;

  END ScrPkT.
