IMPLEMENTATION MODULE Win;
  (*
    Copyright (C) 1993-1995, Alex Iakovlev & Dimon Maslov
    ************************************************************************
    Compilers:  TopSpeed / Stony Brook Modula-2
    ------------------------------------------------------------------------
    A.Iakovlev  Ver  1.0  Derived from Wirth's book              July 87
    A.Iakovlev  Ver  1.1  Use new module RasterDisplay           October 87
    A.Iakovlev  Ver  2.0  Simplifications                        February 89
    A.Iakovlev  Ver  2.1  Corrections in OpenWindow              March 89
    A.Iakovlev  Ver  3.0  New module Geo                         March 89
    A.Iakovlev  Ver  4.0  Init/terminate win.system              March 89
    A.Iakovlev  Ver  5.0  Procedure Inner                        July 89
    A.Iakovlev  Ver  5.1  Correction in Clear                    July 89
    A.Iakovlev  Ver  6.0  Renaming                               January 90
    A.Iakovlev  Ver  7.0  Full redesign                          April 90
    A.Iakovlev  Ver  7.1  New modules RD, *Out                   June 90
    A.Iakovlev  Ver  8.0  No procedures Init/Ready/Term          June 90
    A.Iakovlev  Ver  8.1  Screen pixmap in WinDescriptor         July 90
    A.Iakovlev  Ver  9.0  Simplifications in attributes          July 90
    A.Iakovlev  Ver 10.0  Base point in attributes               August 90
    A.Iakovlev  Ver 10.1  Global screen pixmap                   August 90
    A.Iakovlev  Ver 10.2  New module ScrTxOut                    October 90
    A.Iakovlev  Ver 10.3  Procedure Prev => module SrvObj        January 91
    A.Iakovlev  Ver 10.4  Correction in RestoreWS                January 91
    A.Iakovlev  Ver 11.0  New checks => Status, Dispose changed  February 91
    A.Iakovlev  Ver 12.0  Changes in attributes                  February 91
    A.Iakovlev  Ver 13.0  New procedure Hidden                   February 91
    A.Iakovlev  Ver 14.0  Viewports and clipping                 March 91
    A.Iakovlev  Ver 15.0  Procedures TitleHgt and FrameDims      March 91
    A.Iakovlev  Ver 16.0  Use module Res                         August 92
    A.Iakovlev  Ver 17.0  Pragma data(var_enum_size)             August 92
    A.Iakovlev  Ver 18.0  Procedures Init/Term                   August 92
    A.Iakovlev  Ver 19.0  Renaming                               June 93
    A.Iakovlev  Ver 19.1  Import from SYSTEM                     January 94
    A.Iakovlev  Ver 19.2  Use module Strings                     January 94
    A.Iakovlev  Ver 19.3  Use ScrTxOut/ScrLnOut.CorrectAttrs     February 94
    A.Iakovlev  Ver 19.4  Use module Mem for heap control        October 94
    A.Iakovlev  Ver 19.5  Use new module GeoRgn                  March 95
    A.Iakovlev  Ver 19.6  No procedures Correct*Attrs            March 95
    A.Iakovlev  Ver 20.0  New module ScrOut                      March 95
    ************************************************************************
    TopSpeed compiler dependencies:
    ------ pragmas (*#...*)
    TopSpeed library dependencies:
    ------ import from Strings: Length
    Stony Brook library dependencies:
    ------ import from Strings: Length
  *)

  IMPORT
    SYSTEM, Strings,
    Geo, GeoRgn, Res, SrvObj, DrvMem, Scr, ScrOut;

  (* ======================================== Main constructs *)

  CONST InitClipMode = TRUE;

  TYPE
    WINDOW = POINTER TO WinDescriptor;
    State = (hidden, partially, fully);
    WinDescriptor =
      RECORD
        next: WINDOW;              (* --- next window in window chain *)
        attrs: Attrs;              (* --- window attributes *)
        storePm: Scr.PIXMAP;       (* --- buffer pixmap *)
        innerR: Geo.Rectangle;     (* --- absolute inner rectangle *)
        titleR: Geo.Rectangle;     (* --- absolute title rectangle *)
        outerR: Geo.Rectangle;     (* --- absolute outer rectangle *)
        curView: Scr.View;         (* --- current absolute viewport *)
        oldView: Scr.View;         (* --- old absolute viewport (written by Get) *)
        state: State;              (* --- window state *)
      END;
    Part = (internity, titleOnly, all);
    ChMode = (new, move, size, frame, title);
    ChModeSet = SET OF ChMode;

  CONST
    FrameFont = 0;
  VAR
    screenPm: Scr.PIXMAP;          (* --- screen pixmap *)
    screenR: Geo.Rectangle;        (* --- screen surface rectangle *)
    frameFont: ScrOut.FONT;        (* --- font for text frames *)
    topWin: WINDOW;                (* --- top window in window chain *)
    wasGet: BOOLEAN;               (* --- flag to get surface from Scr *)

  (* ========================================== Initialization/termination *)

  PROCEDURE Init();
    BEGIN
      (* --------- get screen pointer and surface from RD *)
      screenPm := Scr.PagePixmap();
      Scr.Surface(screenPm, screenR);
    END Init;

  PROCEDURE Term();
    VAR
      u: WINDOW;
    BEGIN
      (* -------------------------- dispose all the windows *)
      WHILE topWin # NIL DO
        u := topWin;
        IF Dispose(u) # Res.done THEN
          HALT;
        END;
      END;
    END Term;

  (* ================================= Store/Restore windows *)

  PROCEDURE MakeStore(u: WINDOW; copy: BOOLEAN): BOOLEAN;
  (* make the window store and copy (if copy = TRUE)
     the FULLY VISIBLE window into this store *)
    VAR
      p: Geo.Point;
    BEGIN
      WITH u^ DO
        IF Scr.New(outerR, storePm) = Res.done THEN
          IF copy THEN
            Geo.RectBasePt(outerR, p);
            Scr.RectCopy(storePm, outerR, screenPm, p, Scr.copy);
          END;
          RETURN TRUE;
        ELSE
          RETURN FALSE;
        END;
      END;
    END MakeStore;

  PROCEDURE DropStore(u: WINDOW; copy: BOOLEAN);
  (* copy (if copy = TRUE) the EXISTING window store
     onto the screen WITHOUT CLIPPING and drop the window store *)
    VAR
      p: Geo.Point;
      old: BOOLEAN;
    BEGIN
      WITH u^ DO
        IF copy THEN
          Geo.RectBasePt(outerR, p);
          Scr.SwapClip(screenPm, FALSE, old);
          Scr.RectCopy(screenPm, outerR, storePm, p, Scr.copy);
          Scr.SetClip(screenPm, old);
        END;
        Scr.Dispose(storePm);
      END;
    END DropStore;

  (* ========================================== Show via regions *)

  CONST
    MaxRectNmb = MAX(CARDINAL) DIV SIZE(Geo.Rectangle);

  TYPE
    RectArray = ARRAY [0 .. MaxRectNmb - 1] OF Geo.Rectangle;
    RectArrayPtr = POINTER TO RectArray;

  PROCEDURE ShowStoreRgn(u: WINDOW; rgn: GeoRgn.REGION): BOOLEAN;
  (* show the EXISTING store of the window u via the region rgn
     on the screen WITHOUT CLIPPING  *)
    VAR
      rp: RectArrayPtr;
      n, len, j: CARDINAL;
      p: Geo.Point;
      old: BOOLEAN;
    BEGIN
      n := GeoRgn.RectsNmb(rgn);
      IF n > 0 THEN
        len := n * SIZE(Geo.Rectangle);
        IF DrvMem.Alloc(rp, len) THEN
          GeoRgn.ToRects(rgn, rp^);
          WITH u^ DO
            Scr.SwapClip(screenPm, FALSE, old);
            FOR j := 0 TO n - 1 DO
              Geo.RectBasePt(rp^[j], p);
              Scr.RectCopy(screenPm, rp^[j], storePm, p, Scr.copy);
            END;
            Scr.SetClip(screenPm, old);
          END;
          DrvMem.Free(rp, len);
          RETURN TRUE;
        ELSE
          RETURN FALSE;
        END;
      ELSE
        RETURN TRUE;
      END;
    END ShowStoreRgn;

  PROCEDURE ShowBackgrRgn(rgn: GeoRgn.REGION): BOOLEAN;
  (* show the background via the region rgn
     on the screen WITHOUT CLIPPING *)
    VAR
      rp: RectArrayPtr;
      n, len, j: CARDINAL;
      old: BOOLEAN;
    BEGIN
      n := GeoRgn.RectsNmb(rgn);
      IF n > 0 THEN
        len := n * SIZE(Geo.Rectangle);
        IF DrvMem.Alloc(rp, len) THEN
          GeoRgn.ToRects(rgn, rp^);
          Scr.SwapClip(screenPm, FALSE, old);
          FOR j := 0 TO n - 1 DO
            Scr.SolidRect(screenPm, rp^[j], 0, Scr.copy);
          END;
          Scr.SetClip(screenPm, old);
          DrvMem.Free(rp, len);
          RETURN TRUE;
        ELSE
          RETURN FALSE;
        END;
      ELSE
        RETURN TRUE;
      END;
    END ShowBackgrRgn;

  (* ======================================== Visibility *)

  PROCEDURE TestFullyVis(u: WINDOW): BOOLEAN;
  (* scan the WS and return TRUE, iff the window is fully visible *)
    VAR
      s: WINDOW;
      r: Geo.Rectangle;
    BEGIN
      r := u^.outerR;
      IF Geo.RectInRect(r, screenR) THEN
        (* --------------- window is fully within the screen *)
        s := topWin;
        WHILE s # u DO
          WITH s^ DO
            IF (state # hidden) AND Geo.RectXRect(outerR, r) THEN
              RETURN FALSE;
            END;
          END;
          s := s^.next;
        END;
        RETURN TRUE;
      ELSE
        (* ----------- window is partially or fully out of screen *)
        RETURN FALSE;
      END;
    END TestFullyVis;

  PROCEDURE VisRgnNew(u: WINDOW; up: Part; VAR rgn: GeoRgn.REGION): BOOLEAN;
  (* compute the visibility region rgn of the part up of the window u *)
    VAR
      s: WINDOW;
      r: Geo.Rectangle;
    BEGIN
      (* -------- set the initial rectangle *)
      WITH u^ DO
        CASE up OF
        | internity:
          r := innerR;
        | titleOnly:
          r := titleR;
        | all:
          r := outerR;
        END;
      END;
      (* ----------------------- attempt to create the region *)
      IF GeoRgn.New(r, rgn) THEN
        (* -------------- success: intersect it with screen rectangle *)
        GeoRgn.XRectUpdate(rgn, screenR);
        (* ----------------- minus all the upper windows *)
        s := topWin;
        WHILE s # u DO
          WITH s^ DO
            IF
              (state # hidden) AND
              NOT GeoRgn.MinusRectUpdate(rgn, outerR)
            THEN
              GeoRgn.Dispose(rgn);
              RETURN FALSE;
            END;
          END;
          s := s^.next;
        END;
        RETURN TRUE;
      ELSE
        RETURN FALSE;
      END;
    END VisRgnNew;

  (* =============================================== Show partially *)

  PROCEDURE ShowPartially(u: WINDOW; up: Part): BOOLEAN;
  (* show the part up of the EXISTING PARTIALLY VISIBLE window u
     without changing its layer, position and sizes *)
    VAR
      rgn: GeoRgn.REGION;
      ok: BOOLEAN;
    BEGIN
      (* ----------- compute the visibility region *)
      IF VisRgnNew(u, up, rgn) THEN
        (* ------------- attempt to show the store via the region *)
        ok := ShowStoreRgn(u, rgn);
        GeoRgn.Dispose(rgn);
        RETURN ok;
      ELSE
        (* ------ fail during the computing visibility region *)
        RETURN FALSE;
      END;
    END ShowPartially;

  (* ================================= Store/Restore window systems *)

  PROCEDURE StoreWS(u: WINDOW; VAR r: Geo.Rectangle): BOOLEAN;
  (* recursively store all the windows
     from the window u until end of the window stack,
     marked as fully visible,
     and which have the nonempty intersection with the rectangle r,
     and set their states to partially visible;
     return FALSE, iff such storing failes due to heap limitations:
     in this case WS will be returned to the previous state;
     procedure doesn't any changes on the screen!;
     parameter r is VAR for speed only *)
    BEGIN
      IF u # NIL THEN
        WITH u^ DO
          IF (state = fully) AND Geo.RectXRect(outerR, r) THEN
            (* ------------------------------- attempt to store the window *)
            IF MakeStore(u, TRUE) THEN
              (* ------ success: attempt to process remaining windows *)
              IF StoreWS(next, r) THEN
                (* ----------------- success: that's all *)
                state := partially;
                RETURN TRUE;
              ELSE
                (* ------------ fail: return to the previous state *)
                DropStore(u, FALSE);
                RETURN FALSE;
              END;
            ELSE
              (* ----------- fail during the making store *)
              RETURN FALSE;
            END;
          ELSE
            (* ----------- attempt to process remaining windows *)
            RETURN StoreWS(next, r);
          END;
        END;
      ELSE
        (* --------- end of the window stack *)
        RETURN TRUE;
      END;
    END StoreWS;

  PROCEDURE RestoreWS(u: WINDOW; rgn: GeoRgn.REGION): BOOLEAN;
  (* recursively restore all the windows,
     from the window u until end of the window stack,
     marked as partially visible, via the region rgn,
     and in some cases set their states to fully visible;
     return FALSE, iff such restoring failes due to heap limitations:
     in this case WS will be returned to the previous state,
     but the region MUST be redrawn separately *)
    VAR
      rgnNext, rgnShow: GeoRgn.REGION;
      ok: BOOLEAN;
    BEGIN
      IF GeoRgn.RectsNmb(rgn) > 0 THEN
        IF u # NIL THEN
          WITH u^ DO
            IF (state = partially) AND GeoRgn.XRect(rgn, outerR) THEN
              (* ----------- the window MAY be included in the restoring process *)
              IF TestFullyVis(u) THEN
                (* ------------------ the window will be fully visible *)
                DropStore(u, TRUE);
                (* -------------------------------- attempt to create next region *)
                IF GeoRgn.MinusRectNew(rgn, outerR, rgnNext) THEN
                  (* ------------ success: attempt to process remaining windows *)
                  IF RestoreWS(next, rgnNext) THEN
                    (* ------------- success: that's all *)
                    state := fully;
                    ok := TRUE;
                  ELSE
                    (* ---------------- fail during restoring *)
                    ok := MakeStore(u, TRUE) AND FALSE;
                  END;
                  GeoRgn.Dispose(rgnNext);
                  RETURN ok;
                ELSE
                  (* --------------- fail during region creation *)
                  RETURN MakeStore(u, TRUE) AND FALSE;
                END;
              ELSIF GeoRgn.XRectNew(rgn, outerR, rgnShow) THEN
                (* -------------- partially visible window: attempt to show it *)
                IF ShowStoreRgn(u, rgnShow) THEN
                  (* ---------------------- success: attempt to create next region *)
                  IF GeoRgn.MinusRectNew(rgn, outerR, rgnNext) THEN
                    (* ----------- success: attempt to process remaining windows *)
                    ok := RestoreWS(next, rgnNext);
                    GeoRgn.Dispose(rgnNext);
                  ELSE
                    (* ------- fail during region creation *)
                    ok := FALSE;
                  END;
                ELSE
                  (* ------- fail during showing *)
                  ok := FALSE;
                END;
                GeoRgn.Dispose(rgnShow);
                RETURN ok;
              ELSE
                (* ------- fail during creation of the intersection region *)
                RETURN FALSE;
              END;
            ELSE
              (* ---------------- attempt to process remaining windows *)
              RETURN RestoreWS(next, rgn);
            END;
          END;
        ELSE
          (* -------- end of the window stack: show the background *)
          RETURN ShowBackgrRgn(rgn);
        END;
      ELSE
        (* ---------- the empty region *)
        RETURN TRUE;
      END;
    END RestoreWS;

  (* ================================= Showing/Hiding the windows *)

  PROCEDURE HideShown(u: WINDOW): BOOLEAN;
  (* hide the EXISTING SHOWN window u and update its state;
     return FALSE, iff the showing fails due to heap limitations;
     in this case WS and the screen will be returned to the previous state *)
    VAR
      rgn: GeoRgn.REGION;
      ok: BOOLEAN;
    BEGIN
      WITH u^ DO
        CASE state OF
        | fully:
          (* ------- fully visible onscreen window: attempt to make store *)
          IF MakeStore(u, TRUE) THEN
            (* ----------- success: attempt to create rectangular region *)
            IF GeoRgn.New(outerR, rgn) THEN
              (* --------------- success: attempt to restore WS *)
              state := hidden;
              IF RestoreWS(next, rgn) THEN
                (* ---------- success: that's all! *)
                ok := TRUE;
              ELSE
                (* ------- fail: return the window to the previous state *)
                DropStore(u, TRUE);
                state := fully;
                ok := FALSE;
              END;
              GeoRgn.Dispose(rgn);
              RETURN ok;
            ELSE
              (* ------ fail during creation new region *)
              DropStore(u, FALSE);
              RETURN FALSE;
            END;
          ELSE
            (* ------ fail during making store *)
            RETURN FALSE;
          END;
        | partially:
          (* ---------------- compute the visibility region *)
          IF VisRgnNew(u, all, rgn) THEN
            (* --------- the window visible: attempt to restore WS *)
            state := hidden;
            IF RestoreWS(next, rgn) THEN
              (* ------------- success: that's all! *)
              ok := TRUE;
            ELSE
              (* ------- fail: return the window to the previous state *)
              state := partially;
              ok := ShowStoreRgn(u, rgn) AND FALSE;
            END;
            GeoRgn.Dispose(rgn);
            RETURN ok;
          ELSE
            (* --- fail during the computation of the visibility region *)
            RETURN FALSE;
          END;
        END;
      END;
    END HideShown;

  PROCEDURE ShowHidden(u: WINDOW): BOOLEAN;
  (* show the EXISTING window u and update its state;
     return FALSE, iff the showing fails due to heap limitations;
     in this case WS and the screen will be returned to the previous state *)
    VAR
      rgn: GeoRgn.REGION;
      ok: BOOLEAN;
    BEGIN
      WITH u^ DO
        IF StoreWS(next, outerR) THEN
          (* ------------------------- successful storing of the WS *)
          IF TestFullyVis(u) THEN
            (* ------------ fully visible onscreen window *)
            DropStore(u, TRUE);
            state := fully;
            RETURN TRUE;
          ELSIF VisRgnNew(u, all, rgn) THEN
            (* -------- successful region creation: attempt to show via it *)
            IF ShowStoreRgn(u, rgn) THEN
              (* ----------------- successful showing *)
              state := partially;
              ok := TRUE;
            ELSE
              (* ------------ fail during the showing *)
              ok := FALSE;
            END;
            GeoRgn.Dispose(rgn);
            RETURN ok;
          ELSE
            (* ------------ fail during the region creation *)
            RETURN FALSE;
          END;
        ELSE
          (* --------- fail during the storing WS *)
          RETURN FALSE;
        END;
      END;
    END ShowHidden;

  (* ============================================== Titles *)

  PROCEDURE TitleHgt(a: TitleAttrs): CARDINAL;
    VAR
      w, h: CARDINAL;
    BEGIN
      h := 0;
      CASE a.titleStyle OF
      | titleEmpty:
        (* ------------- no work is needed! *)
      | titleText:
        (* -------------------------------------- text title *)
        ScrOut.Dims(a.titleFont, w, h);
        INC(h, 2 * a.titleTextGap);
      END;
      RETURN h;
    END TitleHgt;

  PROCEDURE ShowTitle(u: WINDOW; VAR a: TitleAttrs; screen: BOOLEAN);
  (* show the title of the window u with attributes a;
     if screen = TRUE, show directly on the screen, otherwise in the memory;
     parameter a is VAR for speed only *)
    VAR
      p: Geo.Point;
      ta: ScrOut.Attrs;
      w, h, slen: CARDINAL;
      j, len: INTEGER;
      pm: Scr.PIXMAP;
    BEGIN
      WITH u^ DO
        (* --------------- set the destination *)
        IF screen THEN
          pm := screenPm;
        ELSE
          pm := storePm;
        END;
        (* -------------------------- draw the title *)
        CASE a.titleStyle OF
        | titleEmpty:
          (* -------- no work is needed *)
        | titleText:
          (* ---------------------------- fill the background *)
          Scr.SolidRect(pm, titleR, a.titleAuxColor, Scr.copy);
          (* ----------------------------------------- draw the text *)
          IF ScrOut.Exist(a.titleFont) THEN
            (* -------------------- find the effective length of string *)
            ScrOut.Dims(a.titleFont, w, h);
            len :=
              (titleR.ixMax - titleR.ixMin + 1 - 2 * VAL(INTEGER, a.titleTextGap))
              DIV VAL(INTEGER, w);
            slen := Strings.Length(a.titleTextStr);
            IF len > VAL(INTEGER, slen) THEN len := slen END;
            (* ----------------------------------------- set attributes *)
            ScrOut.SetTextAttrs
            (
              ta,
              Scr.copy, a.titleColor, 0,
              ScrOut.fillStippled, a.titleFont
            );
            (* ---------------------------------- draw the string *)
            IF len > 0 THEN
              p.ix := titleR.ixMin + VAL(INTEGER, a.titleTextGap);
              p.iy := titleR.iyMin + VAL(INTEGER, a.titleTextGap);
              FOR j := 0 TO len - 1 DO
                ScrOut.Char(pm, ta, p, a.titleTextStr[j]);
                INC(p.ix, w);
              END;
            END;
          END;
        END;
      END;
    END ShowTitle;

  (* ==================================================== Frames *)

  PROCEDURE FrameDims(a: FrameAttrs; VAR w, h: CARDINAL);
    BEGIN
      w := 0;
      h := 0;
      CASE a.frameStyle OF
      | frameEmpty:
        (* -------------- zero dimensions *)
      | frameLine:
        (* ---------------- solid line dimensions *)
        w := 1;
        h := 1;
      | frameText:
        (* -------------------------------- text dimensions *)
        ScrOut.Dims(frameFont, w, h);
      END;
    END FrameDims;

  PROCEDURE ShowFrame(u: WINDOW; VAR a: FrameAttrs; screen: BOOLEAN);
  (* show the frame of the window u with attributes a;
     if screen = TRUE, show directly on the screen, otherwise in the memory;
     parameters a is VAR for speed only *)
    VAR
      p: Geo.Point;
      ta: ScrOut.Attrs;
      w, h: CARDINAL;
      j, wlen, hlen: INTEGER;
      pm: Scr.PIXMAP;
    BEGIN
      WITH u^ DO
        (* --------------- set the destination *)
        IF screen THEN
          pm := screenPm;
        ELSE
          pm := storePm;
        END;
        (* ------------ draw the frame *)
        CASE a.frameStyle OF
        | frameEmpty:
          (* --------------- no work is needed *)
        | frameLine:
          (* --------------------------------- line frame *)
          ScrOut.SetLineAttrs(ta, Scr.copy, a.frameColor);
          ScrOut.Rectangle(pm, ta, outerR);
        | frameText:
          (* ------------------ text frame: find effective lengths *)
          ScrOut.Dims(frameFont, w, h);
          wlen := (outerR.ixMax - outerR.ixMin + 1) DIV VAL(INTEGER, w);
          hlen := (outerR.iyMax - outerR.iyMin + 1) DIV VAL(INTEGER, h);
          (* ------------------------------------ set text attributes *)
          ScrOut.SetTextAttrs
          (
            ta,
            Scr.copy, a.frameColor, a.frameAuxColor,
            ScrOut.fillOpaqueStippled, frameFont
          );
          (* -------------------------------- output: left upper corner *)
          p.ix := outerR.ixMin;
          p.iy := outerR.iyMin;
          ScrOut.Char(pm, ta, p, a.frameTextStr[0]);
          (* --------------------------------------------- upper side *)
          FOR j := 0 TO wlen - 2 DO
            INC(p.ix, w);
            ScrOut.Char(pm, ta, p, a.frameTextStr[1]);
          END;
          (* --------------------------------- right upper corner *)
          p.ix := outerR.ixMax - VAL(INTEGER, w) + 1;
          ScrOut.Char(pm, ta, p, a.frameTextStr[2]);
          (* ---------------------------------------- left side *)
          p.ix := outerR.ixMin;
          FOR j := 0 TO hlen - 2 DO
            INC(p.iy, h);
            ScrOut.Char(pm, ta, p, a.frameTextStr[3]);
          END;
          (* ----------------------------------------- right side *)
          p.ix := outerR.ixMax - VAL(INTEGER, w) + 1;
          p.iy := outerR.iyMin;
          FOR j := 0 TO hlen - 2 DO
            INC(p.iy, h);
            ScrOut.Char(pm, ta, p, a.frameTextStr[4]);
          END;
          (* ------------------------------- left bottom corner *)
          p.ix := outerR.ixMin;
          p.iy := outerR.iyMax - VAL(INTEGER, h) + 1;
          ScrOut.Char(pm, ta, p, a.frameTextStr[5]);
          (* ----------------------------------------- bottom side *)
          FOR j := 0 TO wlen - 2 DO
            INC(p.ix, w);
            ScrOut.Char(pm, ta, p, a.frameTextStr[6]);
          END;
          (* -------------------------------- right bottom corner *)
          p.ix := outerR.ixMax - VAL(INTEGER, w) + 1;
          p.iy := outerR.iyMax - VAL(INTEGER, h) + 1;
          ScrOut.Char(pm, ta, p, a.frameTextStr[7]);
        END;
      END;
    END ShowFrame;

  (* =============================================== Changing windows *)

  PROCEDURE MakeHidden(u: WINDOW; VAR a: Attrs; m: ChMode): BOOLEAN;
  (* make the hidden window u including its store (re)allocation
     according with new attributes a;
     return FALSE, iff such reallocation fails due to heap limitations;
     in this case the window is returned to its previous state;
     parameter a is VAR for speed only *)
    VAR
      d: WinDescriptor;
      copyR: Geo.Rectangle;
      copyP: Geo.Point;
      w, h: CARDINAL;
      ok: BOOLEAN;
    BEGIN
      (* ------ store the old window descriptor on the stack *)
      IF m IN ChModeSet{move, size, frame, title} THEN
        d := u^;
      END;
      (* ---------- set changes in the window descriptor *)
      WITH u^ DO
        (* ----------------------- set attributes *)
        attrs := a;
        (* ------------------- set the inner rectangle *)
        IF m IN ChModeSet{new, move, size} THEN
          innerR.ixMin := 0;
          innerR.ixMax := attrs.wdt - 1;
          innerR.iyMin := 0;
          innerR.iyMax := attrs.hgt - 1;
          Geo.ShiftRect(innerR, attrs.basePt.ix, attrs.basePt.iy);
          curView.port := innerR;
          curView.clip := InitClipMode;
        END;
        (* ---------------------- set the title rectangle *)
        IF m IN ChModeSet{new, move, size, title} THEN
          titleR := innerR;
          titleR.iyMax := titleR.iyMin - 1;
          DEC(titleR.iyMin, TitleHgt(attrs.titleAttrs));
        END;
        (* -------------------- set the outer rectangle *)
        outerR := innerR;
        outerR.iyMin := titleR.iyMin;
        FrameDims(attrs.frameAttrs, w, h);
        DEC(outerR.ixMin, w);
        INC(outerR.ixMax, w);
        DEC(outerR.iyMin, h);
        INC(outerR.iyMax, h);
        (* --------------------- make the new store *)
        storePm := Scr.PIXMAP(NIL);
        IF MakeStore(u, FALSE) THEN
          (* ------------- copy the old memory store onto the new *)
          IF m IN ChModeSet{move, size, frame, title} THEN
            CASE m OF
            | move:
              copyR := outerR;
              Geo.RectBasePt(d.outerR, copyP);
            | size:
              ok := Geo.RectXRectRes(innerR, d.innerR, copyR);
              Geo.RectBasePt(copyR, copyP);
            | frame:
              copyR := innerR;
              copyR.iyMin := titleR.iyMin;
              Geo.RectBasePt(d.titleR, copyP);
            | title:
              copyR := innerR;
              Geo.RectBasePt(d.innerR, copyP);
            END;
            Scr.RectCopy(storePm, copyR, d.storePm, copyP, Scr.copy);
          END;
          (* ------------------------- draw the title *)
          IF m IN ChModeSet{new, size, title} THEN
            ShowTitle(u, attrs.titleAttrs, FALSE);
          END;
          (* ------------------------------ draw the frame *)
          IF m IN ChModeSet{new, size, frame, title} THEN
            ShowFrame(u, attrs.frameAttrs, FALSE);
          END;
          (* ---------------------- drop the old memory store *)
          IF m IN ChModeSet{move, size, frame, title} THEN
            DropStore(SYSTEM.ADR(d), FALSE);
          END;
          (* ---------- normal return *)
          RETURN TRUE;
        ELSE
          (* ------------------ restore the window descriptor *)
          IF m IN ChModeSet{move, size, frame, title} THEN
            u^ := d;
          END;
          (* ----------- fail during making store *)
          RETURN FALSE;
        END;
      END;
    END MakeHidden;

  PROCEDURE ChangeScreen(u: WINDOW; VAR a: Attrs; VAR rgn: GeoRgn.REGION;
    m: ChMode): BOOLEAN;
  (* REALLOCATE the window u according with new attributes a;
     region rgn is equal to the OLD visibility region;
     type of the change is set by parameter m (for speed only!);
     hide the old and show the new window ON THE SCREEN;
     return FALSE, iff such reallocation fails due to heap limitations;
     in this case WS and the screen are returned to their previous state;
     parameter a is VAR for speed only *)
    VAR
      a0: Attrs;
    BEGIN
      WITH u^ DO
        (* ---- store old atributes *)
        a0 := attrs;
        (* ------ attempt to change the window *)
        IF MakeHidden(u, a, m) THEN
          (* -------- success: attempt to modify region for restoring *)
          IF GeoRgn.MinusRectUpdate(rgn, outerR) THEN
            (* --- success: attempt to restore WS and show changed window *)
            IF RestoreWS(next, rgn) AND ShowHidden(u) THEN
              (* --------------------- success: that' all! *)
              RETURN TRUE;
            ELSE
              (* --------------------- fail during the restoring OR the showing *)
              RETURN MakeHidden(u, a0, m) AND ShowStoreRgn(u, rgn) AND FALSE;
            END;
          ELSE
            (* ----------- fail during the region modification *)
            RETURN MakeHidden(u, a0, m) AND FALSE;
          END;
        ELSE
          (* ------- fail during the change *)
          RETURN FALSE;
        END;
      END;
    END ChangeScreen;

  PROCEDURE Change(u: WINDOW; VAR a: Attrs; m: ChMode): BOOLEAN;
  (* change the EXISTING window u according with new attributes a;
     type of the change is set by parameter m (for speed only!);
     return FALSE, iff such reallocation fails due to heap limitations;
     in this case WS and the screen are returned to their previous state;
     parameter a is VAR for speed only *)
    VAR
      rgn: GeoRgn.REGION;
      ok: BOOLEAN;
    BEGIN
      WITH u^ DO
        CASE state OF
        | hidden:
          (* ---------------- the hidden window *)
          RETURN MakeHidden(u, a, m);
        | fully:
          (* ---------------- attempt to make store *)
          IF MakeStore(u, TRUE) THEN
            (* ------- success: attempt to create rectangular region *)
            IF GeoRgn.New(outerR, rgn) THEN
              (* --------------- success: attempt to change *)
              ok := ChangeScreen(u, a, rgn, m);
              GeoRgn.Dispose(rgn);
              RETURN ok;
            ELSE
              (* ------- fail during region creation *)
              DropStore(u, FALSE);
              RETURN FALSE;
            END;
          ELSE
            (* --------- fail during making store *)
            RETURN FALSE;
          END;
        | partially:
          (* --------- attempt to create visibility region *)
          IF VisRgnNew(u, all, rgn) THEN
            (* --------- success: attempt to change *)
            ok := ChangeScreen(u, a, rgn, m);
            GeoRgn.Dispose(rgn);
            RETURN ok;
          ELSE
            (* ------ fail during the region creation *)
            RETURN FALSE;
          END;
        END;
      END;
    END Change;

  (* ==================================== Main window functions *)

  PROCEDURE New(a: Attrs; VAR u: WINDOW): Res.Status;
    VAR
      s: WINDOW;
    BEGIN
      IF (a.wdt > 0) AND (a.hgt > 0) THEN
        IF DrvMem.Alloc(s, SIZE(WinDescriptor)) THEN
          (* ---------------- create rectangles and the store *)
          IF MakeHidden(s, a, new) THEN
            (* --------------------------- set other parameters... *)
            WITH s^ DO
              (* ------------------------------------- ... the state *)
              state := hidden;
              (* ---------- include the window into the window stack *)
              next := topWin;
              topWin := s;
              u := s;
            END;
            (* -------------------------------------- normal return *)
            RETURN Res.done;
          ELSE
            (* --------------------- fail during making the store *)
            DrvMem.Free(s, SIZE(WinDescriptor));
            RETURN Res.noMemory;
          END;
        ELSE
          (* ----------- fail during allocation the descriptor *)
          RETURN Res.noMemory;
        END;
      ELSE
        (* ----------- fail due to bad extent *)
        RETURN Res.badExtent;
      END;
    END New;

  PROCEDURE Show(u: WINDOW): Res.Status;
    BEGIN
      IF u^.state = hidden THEN
        (* ----------- attempt to show the hidden window *)
        IF ShowHidden(u) THEN
          (* ----------------- success: that's all! *)
          RETURN Res.done;
        ELSE
          (* --------------- fail during showing *)
          RETURN Res.noMemory;
        END;
      ELSE
        (* -------- no work is needed! *)
        RETURN Res.done;
      END;
    END Show;

  PROCEDURE Hide(u: WINDOW): Res.Status;
    BEGIN
      IF u^.state # hidden THEN
        (* -------------- attempt to hide the shown window *)
        IF HideShown(u) THEN
          (* -------------- success: that's all! *)
          RETURN Res.done;
        ELSE
          (* -------- fail during the hiding *)
          RETURN Res.noMemory;
        END;
      ELSE
        (* ------- no work is needed! *)
        RETURN Res.done;
      END;
    END Hide;

  PROCEDURE Update(u: WINDOW): Res.Status;
    BEGIN
      IF u^.state = partially THEN
        (* ----------- attempt to show the internity of the window *)
        IF ShowPartially(u, internity) THEN
          (* ------------ success: that's all! *)
          RETURN Res.done;
        ELSE
          (* ----------- fail during the updating *)
          RETURN Res.noMemory;
        END;
      ELSE
        (* ------ no work is needed! *)
        RETURN Res.done;
      END;
    END Update;

  PROCEDURE PutOnTop(u: WINDOW): Res.Status;
    VAR
      prevWin: WINDOW;
    BEGIN
      IF u # topWin THEN
        prevWin := WINDOW(SrvObj.Prev(SrvObj.OBJECT(topWin), SrvObj.OBJECT(u)));
        WITH u^ DO
          (* -------------------------- update the window stack *)
          prevWin^.next := next;
          next := topWin;
          topWin := u;
          (* ----------------------------- update the screen *)
          IF state = partially THEN
            (* ---------------- attempt to show the window *)
            IF ShowHidden(u) THEN
              (* -------- success: that's all! *)
              RETURN Res.done;
            ELSE
              (* ------- fail: return to the previous state *)
              topWin := next;
              next := prevWin^.next;
              prevWin^.next := u;
              RETURN Res.noMemory;
            END;
          ELSE
            (* ---------- no updating is needed! *)
            RETURN Res.done;
          END;
        END;
      ELSE
        (* ----------- no work is needed! *)
        RETURN Res.done;
      END;
    END PutOnTop;

  PROCEDURE Move(u: WINDOW; p: Geo.Point): Res.Status;
    VAR
      a1: Attrs;
    BEGIN
      WITH u^ DO
        (* ------------------------- is the moving necessary? *)
        IF NOT Geo.PtEqPt(p, attrs.basePt) THEN
          (* ------------------- set the new base point *)
          a1 := attrs;
          a1.basePt := p;
          (* ------------- attempt to move the window *)
          IF Change(u, a1, move) THEN
            (* --------- success: that's all! *)
            RETURN Res.done;
          ELSE
            (* ---------- fail during the change *)
            RETURN Res.noMemory;
          END;
        ELSE
          (* ---------- no work is needed! *)
          RETURN Res.done;
        END;
      END;
    END Move;

  PROCEDURE Resize(u: WINDOW; w, h: CARDINAL): Res.Status;
    VAR
      a1: Attrs;
    BEGIN
      IF (w > 0) AND (h > 0) THEN
        WITH u^ DO
          (* ------------------------- is the resizing necessary? *)
          IF (w # attrs.wdt) OR (h # attrs.hgt) THEN
            (* ---------------------- yes: set new attributes *)
            a1 := attrs;
            a1.wdt := w;
            a1.hgt := h;
            (* ----------------------- attempt to resize *)
            IF Change(u, a1, size) THEN
              (* ------------- success: that's all! *)
              RETURN Res.done;
            ELSE
              (* --------- fail during the change *)
              RETURN Res.noMemory;
            END;
          ELSE
            (* ------------- no work is needed! *)
            RETURN Res.done;
          END;
        END;
      ELSE
        (* ----------- fail due to bad extent *)
        RETURN Res.badExtent;
      END;
    END Resize;

  PROCEDURE Reframe(u: WINDOW; a: FrameAttrs): Res.Status;
    VAR
      a1: Attrs;
      w, h: CARDINAL;
    BEGIN
      WITH u^ DO
        (* --------- store attributes *)
        a1 := attrs;
        (* --------------- get dimensions of the new frame *)
        FrameDims(a, w, h);
        (* ---------------------- is the frame simply changeable? *)
        IF (outerR.ixMax - innerR.ixMax = VAL(INTEGER, w)) AND
           (outerR.iyMax - innerR.iyMax = VAL(INTEGER, h)) THEN
          (* ----------- yes: the new frame may be drawn upon the old *)
          attrs.frameAttrs := a;
          CASE state OF
          | hidden:
            ShowFrame(u, a, FALSE);
            RETURN Res.done;
          | partially:
            ShowFrame(u, a, FALSE);
            (* ------------------ attempt to show the new frame *)
            IF ShowPartially(u, all) THEN
              (* ----------- success: that's all! *)
              RETURN Res.done;
            ELSE
              (* ------ fail during the updating: return to the previous *)
              attrs := a1;
              ShowFrame(u, attrs.frameAttrs, FALSE);
              RETURN Res.noMemory;
            END;
          | fully:
            ShowFrame(u, a, TRUE);
            RETURN Res.done;
          END;
        ELSE
          (* ---------- set new attributes *)
          a1.frameAttrs := a;
          (* ------------------ attempt to reframe *)
          IF Change(u, a1, frame) THEN
            (* ---------- success: that's all! *)
            RETURN Res.done;
          ELSE
            (* ------- fail during the change *)
            RETURN Res.noMemory;
          END;
        END;
      END;
    END Reframe;

  PROCEDURE Retitle(u: WINDOW; a: TitleAttrs): Res.Status;
    VAR
      a1: Attrs;
    BEGIN
      WITH u^ DO
        (* --------- get attributes *)
        a1 := attrs;
        (* ----------------------------- is the title simply changeable? *)
        IF titleR.iyMax - titleR.iyMin + 1 = VAL(INTEGER, TitleHgt(a)) THEN
          (* ---------------- yes: the new title may be drawn upon the old *)
          attrs.titleAttrs := a;
          CASE state OF
          | hidden:
            ShowTitle(u, a, FALSE);
            RETURN Res.done;
          | partially:
            ShowTitle(u, a, FALSE);
            (* --------------------- attempt to show the new title *)
            IF ShowPartially(u, titleOnly) THEN
              (* -------------- success: that's all! *)
              RETURN Res.done;
            ELSE
              (* ------ fail during the updating: return to the previous *)
              attrs := a1;
              ShowTitle(u, attrs.titleAttrs, FALSE);
              RETURN Res.noMemory;
            END;
          | fully:
            ShowTitle(u, a, TRUE);
            RETURN Res.done;
          END;
        ELSE
          (* ---------- set new attributes *)
          a1.titleAttrs := a;
          (* -------------------- attempt to retitle *)
          IF Change(u, a1, title) THEN
            (* --------------- that's all! *)
            RETURN Res.done;
          ELSE
            (* ------- fail during the change *)
            RETURN Res.noMemory;
          END;
        END;
      END;
    END Retitle;

  PROCEDURE Dispose(VAR u: WINDOW): Res.Status;
    VAR
      prevWin: WINDOW;
    BEGIN
      prevWin := WINDOW(SrvObj.Prev(SrvObj.OBJECT(topWin), SrvObj.OBJECT(u)));
      IF prevWin # NIL THEN
        IF (u^.state = hidden) OR HideShown(u) THEN
          WITH u^ DO
            (* ------------------------- clear the store *)
            DropStore(u, FALSE);
            (* ------------ exclude the window from the window stack *)
            IF u # topWin THEN
              prevWin^.next := next;
            ELSE
              topWin := next;
            END;
          END;
          (* --------- dispose the window descriptor from the memory *)
          DrvMem.Free(u, SIZE(WinDescriptor));
          RETURN Res.done;
        ELSE
          (* ------------ fail during the window hiding *)
          RETURN Res.noMemory;
        END;
      ELSE
        (* ------------ it isn't window! *)
        RETURN Res.done;
      END;
    END Dispose;

  (* =============================== Viewports and clipping *)

  PROCEDURE Viewport(u: WINDOW; VAR v: Scr.View);
    BEGIN
      WITH u^ DO
        v := curView;
        Geo.ShiftRect(v.port, - innerR.ixMin, - innerR.iyMin);
      END;
    END Viewport;

  PROCEDURE SetViewport(u: WINDOW; v: Scr.View);
    BEGIN
      WITH u^ DO
        Geo.ShiftRect(v.port, innerR.ixMin, innerR.iyMin);
        curView := v;
      END;
    END SetViewport;

  PROCEDURE SwapViewport(u: WINDOW; new: Scr.View; VAR old: Scr.View);
    BEGIN
      WITH u^ DO
        old := curView;
        Geo.ShiftRect(old.port, - innerR.ixMin, - innerR.iyMin);
        Geo.ShiftRect(new.port, innerR.ixMin, innerR.iyMin);
        curView := new;
      END;
    END SwapViewport;

  PROCEDURE Clip(u: WINDOW; VAR c: BOOLEAN);
    BEGIN
      WITH u^ DO
        c := curView.clip;
      END;
    END Clip;

  PROCEDURE SetClip(u: WINDOW; c: BOOLEAN);
    BEGIN
      WITH u^ DO
        curView.clip := c;
      END;
    END SetClip;

  PROCEDURE SwapClip(u: WINDOW; new: BOOLEAN; VAR old: BOOLEAN);
    BEGIN
      WITH u^.curView DO
        old := clip;
        clip := new;
      END;
    END SwapClip;

  (* ================================ Inquery functions *)

  PROCEDURE Exist(u: WINDOW): BOOLEAN;
    BEGIN
      RETURN SrvObj.Exist(SrvObj.OBJECT(topWin), SrvObj.OBJECT(u));
    END Exist;

  PROCEDURE FullyVisible(u: WINDOW): BOOLEAN;
    BEGIN
      RETURN u^.state = fully;
    END FullyVisible;

  PROCEDURE Hidden(u: WINDOW): BOOLEAN;
    BEGIN
      RETURN u^.state = hidden;
    END Hidden;

  PROCEDURE Top(): WINDOW;
    BEGIN
      RETURN topWin;
    END Top;

  PROCEDURE Next(u: WINDOW): WINDOW;
    BEGIN
      RETURN u^.next;
    END Next;

  PROCEDURE Find(pt: Geo.Point): WINDOW;
    VAR
      s: WINDOW;
    BEGIN
      s := topWin;
      LOOP
        IF s = NIL THEN
          RETURN NIL;
        ELSIF (s^.state # hidden) AND Geo.PtInRect(pt, s^.outerR) THEN
          RETURN s;
        ELSE
          s := s^.next;
        END;
      END;
    END Find;

  PROCEDURE Attributes(u: WINDOW; VAR a: Attrs);
    BEGIN
      a := u^.attrs;
    END Attributes;

  (* ================================== Inland functions *)

  PROCEDURE Get(u: WINDOW; VAR i: Inland);
    BEGIN
      WITH u^ DO
        (* ----------- fill inland *)
        CASE state OF
        | fully:
          i.pixmap := screenPm;
        | partially, hidden:
          i.pixmap := storePm;
        END;
        i.port := curView.port;
        i.org.ix := innerR.ixMin;
        i.org.iy := innerR.iyMin;
        (* -------------------- change the clipping state of pixmap *)
        Scr.SwapViewport(i.pixmap, curView, oldView);
      END;
    END Get;

  PROCEDURE Release(u: WINDOW);
    BEGIN
      WITH u^ DO
        CASE state OF
        | fully:
          Scr.SetViewport(screenPm, oldView);
        | partially, hidden:
          Scr.SetViewport(storePm, oldView);
        END;
      END;
    END Release;

  BEGIN
    topWin := NIL;
    frameFont := ScrOut.StdFont(FrameFont);
  END Win.
